# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-locals
# pylint: disable=too-many-lines

"""Manages the speech engine: server, synthesizer, voice, and output parameters."""

from __future__ import annotations

import importlib
import os
from typing import TYPE_CHECKING, Any

from . import (
    command_manager,
    dbus_service,
    debug,
    gsettings_registry,
    guilabels,
    input_event,
    language_utilities,
    messages,
    presentation_manager,
    speech_manager_command_definitions,
    speechserver,
    systemd,
)
from .acss import ACSS
from .extension import Extension
from .speechserver import CapitalizationStyle, PunctuationStyle

if TYPE_CHECKING:
    from .command import Command, KeyboardCommand
    from .dbus_service import UInt32
    from .scripts import default
    from .speech_manager_preferences_grid import VoicesPreferencesGrid, VoiceTypesPreferencesGrid
    from .speechserver import SpeechServer

SPEECH_FACTORY_MODULES: list[str] = ["speechdispatcherfactory", "spiel"]


@gsettings_registry.get_registry().gsettings_schema("org.gnome.Orca.Speech", name="speech")
@gsettings_registry.get_registry().gsettings_schema("org.gnome.Orca.Voice", name="voice")
class SpeechManager(Extension):
    """Manages the speech engine: server, synthesizer, voice, and output parameters."""

    GROUP_LABEL = guilabels.KB_GROUP_SPEECH_VERBOSITY

    SPEECH_SCHEMA = "speech"
    _VOICE_SCHEMA = "voice"

    KEY_ENABLE = "enable"
    KEY_SPEECH_SERVER = "speech-server"
    KEY_SPEECH_SERVER_FACTORY = "speech-server-factory"
    KEY_SYNTHESIZER = "synthesizer"
    KEY_SPEAK_NUMBERS_AS_DIGITS = "speak-numbers-as-digits"
    KEY_USE_COLOR_NAMES = "use-color-names"
    KEY_INSERT_PAUSES_BETWEEN_UTTERANCES = "insert-pauses-between-utterances"
    KEY_USE_PRONUNCIATION_DICTIONARY = "use-pronunciation-dictionary"
    KEY_AUTO_LANGUAGE_SWITCHING = "auto-language-switching"
    KEY_AUTO_LANGUAGE_SWITCHING_UI = "auto-language-switching-ui"
    KEY_ONLY_SWITCH_CONFIGURED_LANGUAGES = "only-switch-configured-languages"
    KEY_CAPITALIZATION_STYLE = "capitalization-style"
    KEY_PUNCTUATION_LEVEL = "punctuation-level"

    KEY_ESTABLISHED = "established"
    KEY_RATE = "rate"
    KEY_PITCH = "pitch"
    KEY_PITCH_RANGE = "pitch-range"
    KEY_VOLUME = "volume"
    KEY_FAMILY_NAME = "family-name"
    KEY_FAMILY_LANG = "family-lang"
    KEY_FAMILY_DIALECT = "family-dialect"
    KEY_FAMILY_GENDER = "family-gender"
    KEY_FAMILY_VARIANT = "family-variant"

    def _get_setting(
        self,
        key: str,
        gtype: str,
        default: Any,
        app_name: str | None = None,
        genum: str | None = None,
    ) -> Any:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self.SPEECH_SCHEMA,
            key,
            gtype,
            genum=genum,
            default=default,
            app_name=app_name,
        )

    def get_voice_properties(
        self,
        voice_type: str = "",
        app_name: str | None = None,
        voice_set: str = gsettings_registry.PRIMARY_VOICE_SET,
    ) -> ACSS:
        """Returns voice properties from dconf for the given voice type and set."""

        vtype = voice_type or speechserver.VoiceType.DEFAULT

        if voice_set != gsettings_registry.PRIMARY_VOICE_SET:
            return self._get_voice_set_properties(vtype, voice_set)

        lookup = gsettings_registry.get_registry().layered_lookup
        voice: dict[str, Any] = {}

        established = lookup(
            self._VOICE_SCHEMA,
            self.KEY_ESTABLISHED,
            "b",
            voice_type=vtype,
            app_name=app_name,
        )
        if established is not None:
            voice["established"] = established

        rate = lookup(self._VOICE_SCHEMA, self.KEY_RATE, "i", voice_type=vtype, app_name=app_name)
        if rate is not None:
            voice[ACSS.RATE] = rate
        pitch = lookup(self._VOICE_SCHEMA, self.KEY_PITCH, "d", voice_type=vtype, app_name=app_name)
        if pitch is not None:
            voice[ACSS.AVERAGE_PITCH] = pitch
        pitch_range = lookup(
            self._VOICE_SCHEMA, self.KEY_PITCH_RANGE, "d", voice_type=vtype, app_name=app_name
        )
        if pitch_range is not None:
            voice[ACSS.PITCH_RANGE] = pitch_range
        volume = lookup(
            self._VOICE_SCHEMA, self.KEY_VOLUME, "d", voice_type=vtype, app_name=app_name
        )
        if volume is not None:
            voice[ACSS.GAIN] = volume

        family: dict[str, str] = {}
        for dconf_key, family_key in (
            (self.KEY_FAMILY_NAME, speechserver.VoiceFamily.NAME),
            (self.KEY_FAMILY_LANG, speechserver.VoiceFamily.LANG),
            (self.KEY_FAMILY_DIALECT, speechserver.VoiceFamily.DIALECT),
            (self.KEY_FAMILY_GENDER, speechserver.VoiceFamily.GENDER),
            (self.KEY_FAMILY_VARIANT, speechserver.VoiceFamily.VARIANT),
        ):
            value = lookup(self._VOICE_SCHEMA, dconf_key, "s", voice_type=vtype, app_name=app_name)
            if value:
                family[family_key] = value
        if family:
            voice[ACSS.FAMILY] = family

        return ACSS(voice)

    def get_voice_set_voice(
        self, voice_type: str, voice_set: str, fall_back_to_default: bool = True
    ) -> ACSS:
        """Returns the set's voice for voice_type, optionally falling back to its default voice."""

        if fall_back_to_default:
            types: tuple[str, ...] = (voice_type, speechserver.VoiceType.DEFAULT)
        else:
            types = (voice_type,)
        for vtype in types:
            config = self.get_voice_properties(vtype, voice_set=voice_set)
            if config and config.get(self.KEY_ESTABLISHED):
                return config
        return ACSS({})

    def _set_runtime_voice_family(self, voice_type: str, family: dict[str, str]) -> None:
        """Overrides the runtime voice family for voice_type, clearing any previous override."""

        registry = gsettings_registry.get_registry()
        for family_key, key in (
            (speechserver.VoiceFamily.NAME, self.KEY_FAMILY_NAME),
            (speechserver.VoiceFamily.LANG, self.KEY_FAMILY_LANG),
            (speechserver.VoiceFamily.DIALECT, self.KEY_FAMILY_DIALECT),
            (speechserver.VoiceFamily.GENDER, self.KEY_FAMILY_GENDER),
            (speechserver.VoiceFamily.VARIANT, self.KEY_FAMILY_VARIANT),
        ):
            registry.remove_runtime_value(self._VOICE_SCHEMA, key, voice_type=voice_type)
            if value := family.get(family_key):
                registry.set_runtime_value(self._VOICE_SCHEMA, key, value, voice_type=voice_type)

    def _set_runtime_voice(self, voice_type: str, voice: ACSS) -> None:
        """Overrides the runtime voice for voice_type, clearing any previous override."""

        registry = gsettings_registry.get_registry()
        prosody = (
            (ACSS.RATE, self.KEY_RATE),
            (ACSS.AVERAGE_PITCH, self.KEY_PITCH),
            (ACSS.PITCH_RANGE, self.KEY_PITCH_RANGE),
            (ACSS.GAIN, self.KEY_VOLUME),
        )
        for _attr, key in ((None, self.KEY_ESTABLISHED), *prosody):
            registry.remove_runtime_value(self._VOICE_SCHEMA, key, voice_type=voice_type)
        # A non-default voice type is only used if it is established; mirror that here so the
        # generator applies it rather than falling back to the default voice.
        registry.set_runtime_value(
            self._VOICE_SCHEMA,
            self.KEY_ESTABLISHED,
            bool(voice.get("established")),
            voice_type=voice_type,
        )
        for acss_key, key in prosody:
            if acss_key in voice:
                registry.set_runtime_value(
                    self._VOICE_SCHEMA, key, voice[acss_key], voice_type=voice_type
                )
        self._set_runtime_voice_family(voice_type, voice.get(ACSS.FAMILY, {}))

    def apply_live_voice(self, voice: ACSS) -> None:
        """Auditions voice as the live default voice, bypassing voice-set rerouting."""

        self._auditioning_voice = True
        self._set_runtime_voice(speechserver.VoiceType.DEFAULT, voice)
        if (server := self.get_server()) is not None:
            server.set_default_voice(voice)
            server.clear_cached_voice_properties()

    def restore_live_voices(self, voices: dict[str, ACSS]) -> None:
        """Makes each staged voice live by voice type and resumes normal voice-set processing."""

        self._auditioning_voice = False
        for voice_type, voice in voices.items():
            self._set_runtime_voice(voice_type, voice)
        if (server := self.get_server()) is not None:
            server.set_default_voice(voices.get(speechserver.VoiceType.DEFAULT, ACSS({})))
            server.clear_cached_voice_properties()

    @staticmethod
    def apply_voice_overrides(base: ACSS, override: ACSS) -> ACSS:
        """Overlays family and prosody from override onto base, returning the merged ACSS."""

        tokens = ["SPEECH MANAGER: Applying voice overrides:", override, "onto:", base]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if ACSS.FAMILY in override:
            family = dict(base.get(ACSS.FAMILY, {}))
            family.update(override[ACSS.FAMILY])
            base[ACSS.FAMILY] = family
        for prop in (ACSS.RATE, ACSS.AVERAGE_PITCH, ACSS.PITCH_RANGE, ACSS.GAIN):
            if prop in override:
                base[prop] = override[prop]
        return base

    def apply_voice_set(self, voice: ACSS) -> ACSS:
        """Overlays the active voice set, or the set matching the voice's language."""

        if self._auditioning_voice:
            # A voice is being auditioned in preferences; let it be heard as configured.
            return voice
        voice_type = voice.get(ACSS.VOICE_TYPE, speechserver.VoiceType.DEFAULT)
        fall_back_to_default = True
        if self._active_voice_set != gsettings_registry.PRIMARY_VOICE_SET:
            # A command loaded a set manually; it overrides every voice type. The system
            # voice is used only if the set configures one; otherwise status messages stay
            # in Orca's own voice rather than borrowing the set's default voice.
            voice_set = self._active_voice_set
            fall_back_to_default = voice_type != speechserver.VoiceType.SYSTEM
        else:
            # Otherwise follow the voice's own language (automatic switching).
            family = voice.get(ACSS.FAMILY)
            if not family:
                return voice
            lang = family.get(speechserver.VoiceFamily.LANG, "")
            if not lang:
                return voice
            dialect = family.get(speechserver.VoiceFamily.DIALECT, "")
            voice_set = f"{lang}-{dialect}".lower() if dialect else lang.lower()
            voice_set_names = self.get_voice_set_names()
            if voice_set not in voice_set_names:
                voice_set = lang.lower()
            if voice_set not in voice_set_names:
                return voice

        voice.pop(ACSS.VOICE_TYPE, None)
        config = self.get_voice_set_voice(voice_type, voice_set, fall_back_to_default)
        if not config:
            return voice

        tokens = ["SPEECH MANAGER: Applying voice set", voice_set, "for", voice_type]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return self.apply_voice_overrides(voice, config)

    def _get_voice_set_properties(self, voice_type: str, voice_set: str) -> ACSS:
        """Returns voice properties for a non-primary voice set."""

        registry = gsettings_registry.get_registry()
        sub = gsettings_registry.get_registry().voice_set_sub_path(voice_type, voice_set)
        gs = registry.get_settings("voice", registry.get_active_profile(), sub)
        if gs is None:
            return ACSS({})

        voice: dict[str, Any] = {}
        for key, acss_key in (
            (self.KEY_RATE, ACSS.RATE),
            (self.KEY_PITCH, ACSS.AVERAGE_PITCH),
            (self.KEY_PITCH_RANGE, ACSS.PITCH_RANGE),
            (self.KEY_VOLUME, ACSS.GAIN),
        ):
            if gs.get_user_value(key) is not None:
                voice[acss_key] = gs.get_value(key).unpack()

        established = gs.get_user_value(self.KEY_ESTABLISHED)
        if established is not None:
            voice["established"] = established.get_boolean()

        family: dict[str, str] = {}
        for dconf_key, family_key in (
            (self.KEY_FAMILY_NAME, speechserver.VoiceFamily.NAME),
            (self.KEY_FAMILY_LANG, speechserver.VoiceFamily.LANG),
            (self.KEY_FAMILY_DIALECT, speechserver.VoiceFamily.DIALECT),
            (self.KEY_FAMILY_GENDER, speechserver.VoiceFamily.GENDER),
            (self.KEY_FAMILY_VARIANT, speechserver.VoiceFamily.VARIANT),
        ):
            val = gs.get_user_value(dconf_key)
            if val is not None:
                family[family_key] = val.get_string()
        if family:
            voice[ACSS.FAMILY] = family

        return ACSS(voice)

    def set_voice_set_properties(self, voice_type: str, voice_set: str, properties: ACSS) -> None:
        """Stores voice properties for a voice set."""

        registry = gsettings_registry.get_registry()
        sub = gsettings_registry.get_registry().voice_set_sub_path(voice_type, voice_set)
        gs = registry.get_settings("voice", registry.get_active_profile(), sub)
        if gs is None:
            return

        gs.set_boolean(self.KEY_ESTABLISHED, True)

        for key, acss_key, setter in (
            (self.KEY_RATE, ACSS.RATE, gs.set_int),
            (self.KEY_PITCH, ACSS.AVERAGE_PITCH, gs.set_double),
            (self.KEY_PITCH_RANGE, ACSS.PITCH_RANGE, gs.set_double),
            (self.KEY_VOLUME, ACSS.GAIN, gs.set_double),
        ):
            if acss_key in properties:
                setter(key, properties[acss_key])

        family = properties.get(ACSS.FAMILY, {})
        for dconf_key, family_key in (
            (self.KEY_FAMILY_NAME, speechserver.VoiceFamily.NAME),
            (self.KEY_FAMILY_LANG, speechserver.VoiceFamily.LANG),
            (self.KEY_FAMILY_DIALECT, speechserver.VoiceFamily.DIALECT),
            (self.KEY_FAMILY_GENDER, speechserver.VoiceFamily.GENDER),
            (self.KEY_FAMILY_VARIANT, speechserver.VoiceFamily.VARIANT),
        ):
            if family.get(family_key):
                gs.set_string(dconf_key, family[family_key])

    def get_voice_set_names(self) -> list[str]:
        """Returns the names of configured voice sets (excluding primary)."""

        registry = gsettings_registry.get_registry()
        profile = registry.get_active_profile()
        gs = registry.get_settings(
            "voice", profile, f"voice-sets/{gsettings_registry.PRIMARY_VOICE_SET}/default"
        )
        if gs is None:
            return []

        entries = gsettings_registry.GSettingsRegistry.dconf_list(
            f"{gsettings_registry.GSETTINGS_PATH_PREFIX}"
            f"{gsettings_registry.GSettingsRegistry.sanitize_gsettings_path(profile)}/voice-sets/"
        )
        return [e for e in entries if e != gsettings_registry.PRIMARY_VOICE_SET]

    def __init__(self) -> None:
        self._families_sorted: bool = False
        self._mute_speech: bool = False
        self._server: SpeechServer | None = None
        self._active_voice_set: str = gsettings_registry.PRIMARY_VOICE_SET
        self._auditioning_voice = False
        self._voice_set_command_names: set[str] = set()
        super().__init__()
        gsettings_registry.get_registry().add_profile_change_observer(
            self.refresh_voice_set_commands
        )

    def _get_commands(self) -> list[Command]:
        """Returns commands for registration."""

        commands = speech_manager_command_definitions.get_commands(self)
        commands.extend(self._build_voice_set_commands())
        return commands

    def _build_voice_set_commands(self) -> list[KeyboardCommand]:
        """Returns one unbound activation command per available voice set."""

        commands = speech_manager_command_definitions.get_voice_set_commands(self)
        self._voice_set_command_names = {cmd.get_name() for cmd in commands}
        return commands

    def refresh_voice_set_commands(self, _profile: str = "") -> None:
        """Re-registers the per-set activation commands after the available sets change."""

        if not self._commands_initialized:
            return

        manager = command_manager.get_manager()
        previous = self._voice_set_command_names
        commands = self._build_voice_set_commands()
        for name in previous:
            manager.remove_command(name)
        for command in commands:
            manager.add_command(command)

    def get_server(self) -> SpeechServer | None:
        """Returns the speech server instance, or None if not initialized."""

        return self._server

    def _get_server(self) -> SpeechServer | None:
        """Returns the speech server if it is responsive."""

        result = self._server
        if result is None:
            msg = "SPEECH MANAGER: Speech server is None."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return None

        if not result.is_responsive():
            msg = "SPEECH MANAGER: Speech server is not responsive."
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return None

        tokens = ["SPEECH MANAGER: Speech server is", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    def _get_available_servers(self) -> list[str]:
        """Returns a list of available speech servers."""

        return list(self._get_server_module_map().keys())

    def _get_server_module_map(self) -> dict[str, str]:
        """Returns a mapping of server names to module names."""

        result = {}
        for module_name in SPEECH_FACTORY_MODULES:
            try:
                factory = importlib.import_module(f"orca.{module_name}")
            except ImportError:
                try:
                    factory = importlib.import_module(module_name)
                except ImportError:
                    continue

            try:
                speech_server_class = factory.SpeechServer
                if server_name := speech_server_class.get_factory_name():
                    result[server_name] = module_name

            except (AttributeError, TypeError, ImportError) as error:
                tokens = [f"SPEECH MANAGER: {module_name} not available:", error]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return result

    def _switch_server(self, target_server: str) -> bool:
        """Switches to the specified server."""

        server_module_map = self._get_server_module_map()
        target_module = server_module_map.get(target_server)
        if not target_module:
            return False

        self.shutdown_speech()
        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA,
            self.KEY_SPEECH_SERVER_FACTORY,
            target_module,
        )
        self.start_speech()
        return self.get_current_server() == target_server

    @dbus_service.getter
    def get_available_servers(self) -> list[str]:
        """Returns a list of available servers."""

        result = self._get_available_servers()
        msg = f"SPEECH MANAGER: Available servers: {result}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SPEECH_SERVER,
        schema="speech",
        gtype="s",
        default="",
        summary="Speech server name",
    )
    @dbus_service.getter
    def get_current_server(self) -> str:
        """Returns the name of the current speech server (Speech Dispatcher or Spiel)."""

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return ""

        name = server.get_factory_name()
        msg = f"SPEECH MANAGER: Server is: {name}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return name

    @dbus_service.setter
    def set_current_server(self, value: str) -> bool:
        """Sets the current speech server (e.g. Speech Dispatcher or Spiel)."""

        return self._switch_server(value)

    def get_speech_server(self, app_name: str | None = None) -> str:
        """Returns the speech server setting."""

        return self._get_setting(self.KEY_SPEECH_SERVER, "s", "", app_name=app_name)

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SPEECH_SERVER_FACTORY,
        schema="speech",
        gtype="s",
        default="speechdispatcherfactory",
        summary="Speech server factory module",
        migration_key="speechServerFactory",
    )
    def get_speech_server_factory(self) -> str:
        """Returns the speech server factory module name."""

        # Test override: the integration test harness points this at a no-audio server.
        if override := os.environ.get("ORCA_TEST_SPEECH_SERVER_FACTORY"):
            return override
        return self._get_setting(self.KEY_SPEECH_SERVER_FACTORY, "s", "speechdispatcherfactory")

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SYNTHESIZER,
        schema="speech",
        gtype="s",
        default="",
        summary="Speech synthesizer",
    )
    @dbus_service.getter
    def get_current_synthesizer(self) -> str:
        """Returns the current synthesizer of the speech server."""

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return ""

        result = server.get_output_module()
        msg = f"SPEECH MANAGER: Synthesizer is: {result}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    @dbus_service.setter
    def set_current_synthesizer(self, value: str) -> bool:
        """Sets the current synthesizer of the active speech server."""

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        available = self.get_available_synthesizers()
        if value not in available:
            tokens = [f"SPEECH MANAGER: '{value}' is not in", available]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        msg = f"SPEECH MANAGER: Setting synthesizer to: {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        server.set_output_module(value)
        return server.get_output_module() == value

    def get_synthesizer(self, app_name: str | None = None) -> str:
        """Returns the synthesizer setting."""

        return self._get_setting(self.KEY_SYNTHESIZER, "s", "", app_name=app_name)

    @dbus_service.getter
    def get_available_synthesizers(self) -> list[str]:
        """Returns a list of available synthesizers of the speech server."""

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return []

        synthesizers = server.get_speech_servers()
        result = [s.get_info()[1] for s in synthesizers]
        msg = f"SPEECH MANAGER: Available synthesizers: {result}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    @dbus_service.getter
    def get_available_voices(self) -> list[str]:
        """Returns a list of available voices for the current synthesizer."""

        server = self._get_server()
        if server is None:
            return []

        voices = server.get_voice_families()
        if not voices:
            return []

        result = sorted(
            {
                voice_name
                for voice in voices
                if (voice_name := voice.get(speechserver.VoiceFamily.NAME, ""))
            },
        )
        return result

    def get_voice_families(self) -> list[speechserver.VoiceFamily]:
        """Returns the full list of voice family dictionaries for the current synthesizer.
        Each dictionary contains NAME, LANG, DIALECT, and VARIANT fields."""

        server = self._get_server()
        if server is None:
            return []

        return server.get_voice_families() or []

    @dbus_service.parameterized_command
    def get_voices_for_language(
        self,
        language: str,
        variant: str = "",
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> list[tuple[str, str, str]]:
        """Returns a list of available voices for the specified language."""

        tokens = [
            "SPEECH MANAGER: get_voices_for_language. Language:",
            language,
            "Variant:",
            variant,
            "Script:",
            script,
            "Event:",
            event,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        server = self._get_server()
        if server is None:
            return []

        voices = server.get_voice_families_for_language(language=language, variant=variant)
        result = []
        for name, lang, var in voices:
            result.append((name, lang or "", var or ""))

        msg = f"SPEECH MANAGER: Found {len(result)} voice(s) for '{language}'."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FAMILY_NAME,
        schema="voice",
        gtype="s",
        default="",
        summary="Voice family name",
    )
    @dbus_service.getter
    def get_current_voice(self) -> str:
        """Returns the current voice name."""

        server = self._get_server()
        if server is None:
            return ""

        result = ""
        if voice_family := server.get_voice_family():
            result = voice_family.get(speechserver.VoiceFamily.NAME, "")

        return result

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FAMILY_LANG,
        schema="voice",
        gtype="s",
        default="",
        summary="Voice family language",
    )
    def get_current_voice_lang(self) -> str:
        """Returns the language of the current voice."""

        server = self._get_server()
        if server is None:
            return ""

        if voice_family := server.get_voice_family():
            return voice_family.get(speechserver.VoiceFamily.LANG, "") or ""

        return ""

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FAMILY_DIALECT,
        schema="voice",
        gtype="s",
        default="",
        summary="Voice family dialect",
    )
    def get_current_voice_dialect(self) -> str:
        """Returns the dialect of the current voice."""

        server = self._get_server()
        if server is None:
            return ""

        if voice_family := server.get_voice_family():
            return voice_family.get(speechserver.VoiceFamily.DIALECT, "") or ""

        return ""

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FAMILY_GENDER,
        schema="voice",
        gtype="s",
        default="",
        summary="Voice family gender",
    )
    def get_current_voice_gender(self) -> str:
        """Returns the gender of the current voice."""

        server = self._get_server()
        if server is None:
            return ""

        if voice_family := server.get_voice_family():
            return voice_family.get(speechserver.VoiceFamily.GENDER, "") or ""

        return ""

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FAMILY_VARIANT,
        schema="voice",
        gtype="s",
        default="",
        summary="Voice family variant",
    )
    def get_current_voice_variant(self) -> str:
        """Returns the variant of the current voice."""

        server = self._get_server()
        if server is None:
            return ""

        if voice_family := server.get_voice_family():
            return voice_family.get(speechserver.VoiceFamily.VARIANT, "") or ""

        return ""

    @dbus_service.setter
    def set_current_voice(self, voice_name: str) -> bool:
        """Sets the current voice for the active synthesizer."""

        server = self._get_server()
        if server is None:
            return False

        available = self.get_available_voices()
        if voice_name not in available:
            msg = f"SPEECH MANAGER: '{voice_name}' is not in {available}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        voices = server.get_voice_families()
        if not voices:
            return False

        result = False
        for voice_family in voices:
            family_name = voice_family.get(speechserver.VoiceFamily.NAME, "")
            if family_name == voice_name:
                server.set_voice_family(voice_family)
                self._set_runtime_voice_family(speechserver.VoiceType.DEFAULT, voice_family)
                result = True
                break

        msg = f"SPEECH MANAGER: Set voice to '{voice_name}': {result}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def get_current_speech_server_info(self) -> tuple[str, str]:
        """Returns the name and ID of the current speech server."""

        # TODO - JD: The result is not in sync with the current output module. Should it be?
        # TODO - JD: The only caller is the preferences dialog. And the useful functionality is in
        # the methods to get (and set) the output module. So why exactly do we need this?
        server = self._get_server()
        if server is None:
            return ("", "")

        server_name, server_id = server.get_info()
        msg = f"SPEECH MANAGER: Speech server info: {server_name}, {server_id}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return server_name, server_id

    def check_speech_setting(self) -> None:
        """Checks the speech setting and initializes speech if necessary."""

        if not self.get_speech_is_enabled():
            msg = "SPEECH MANAGER: Speech is not enabled. Shutting down speech."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.shutdown_speech()
            return

        msg = "SPEECH MANAGER: Speech is enabled."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self.start_speech()

    @dbus_service.command
    def start_speech(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = False,
    ) -> bool:
        """Starts the speech server."""

        tokens = [
            "SPEECH MANAGER: start_speech. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        self._init_server()
        return True

    def _init_server(self) -> None:
        """Initializes the speech server."""

        debug.print_message(debug.LEVEL_INFO, "SPEECH MANAGER: Initializing server", True)
        if self._server:
            debug.print_message(debug.LEVEL_INFO, "SPEECH MANAGER: Already initialized", True)
            return

        factory = self.get_speech_server_factory()
        self._server = self._init_server_from_module(factory, None)

        if not self._server:
            for module_name in SPEECH_FACTORY_MODULES:
                if module_name != factory:
                    self._server = self._init_server_from_module(module_name, None)
                    if self._server:
                        factory = module_name
                        break

        if self._server:
            tokens = ["SPEECH MANAGER: Using speech server factory:", factory]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

            synth = self._get_setting(self.KEY_SYNTHESIZER, "s", "")
            if synth:
                self._server.set_output_module(synth)

            self._server.set_default_voice(self.get_voice_properties())
            level_str = self.get_punctuation_level()
            self._server.update_punctuation_level(PunctuationStyle[level_str.upper()])
            self._server.update_capitalization_style(self.get_capitalization_style())
        else:
            msg = "SPEECH MANAGER: Speech not available"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            systemd.get_manager().set_status("Speech", "not available")

        debug.print_message(debug.LEVEL_INFO, "SPEECH MANAGER: Server initialized", True)
        if self._server:
            systemd.get_manager().set_status("Speech", "enabled")

    @staticmethod
    def _init_server_from_module(
        module_name: str,
        speech_server_info: list[str] | None,
    ) -> SpeechServer | None:
        """Attempts to initialize a speech server from the given module."""

        if not module_name:
            return None

        factory = None
        try:
            factory = importlib.import_module(f"orca.{module_name}")
        except ImportError:
            try:
                factory = importlib.import_module(module_name)
            except ImportError:
                debug.print_exception(debug.LEVEL_SEVERE)

        if not factory:
            msg = f"SPEECH MANAGER: Failed to import module: {module_name}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return None

        server = None
        if speech_server_info:
            server = factory.SpeechServer.get_speech_server(speech_server_info)

        if not server:
            if speech_server_info:
                tokens = ["SPEECH MANAGER: Could not use server info:", speech_server_info]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            server = factory.SpeechServer.get_speech_server()

        if not server:
            msg = f"SPEECH MANAGER: No speech server for factory: {module_name}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)

        return server

    @dbus_service.command
    def interrupt_speech(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = False,
    ) -> bool:
        """Interrupts the speech server."""

        tokens = [
            "SPEECH MANAGER: interrupt_speech. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if server := self._get_server():
            server.stop()

        return True

    @dbus_service.command
    def shutdown_speech(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = False,
    ) -> bool:
        """Shuts down the speech server."""

        tokens = [
            "SPEECH MANAGER: shutdown_speech. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if server := self._get_server():
            server.shutdown_active_servers()
            self._server = None

        return True

    @dbus_service.command
    def refresh_speech(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = False,
    ) -> bool:
        """Shuts down and re-initializes speech."""

        tokens = [
            "SPEECH MANAGER: refresh_speech. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self.shutdown_speech()
        self.start_speech()
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_ESTABLISHED,
        schema="voice",
        gtype="b",
        default=False,
        migration_key="established",
        summary="Whether this voice type has been user-customized",
    )
    def get_established(self) -> bool:
        """Returns whether the current voice type has been customized."""

        return False

    @gsettings_registry.get_registry().gsetting(
        key=KEY_RATE,
        schema="voice",
        gtype="i",
        default=50,
        summary="Speech rate (0-100)",
    )
    @dbus_service.getter
    def get_rate(self) -> UInt32:
        """Returns the current speech rate."""

        return gsettings_registry.get_registry().layered_lookup(
            self._VOICE_SCHEMA,
            self.KEY_RATE,
            "i",
            default=50,
        )

    def _sync_runtime_value_to_all_voice_types(self, key: str, value: Any) -> None:
        """Sets a runtime value override for all voice types."""

        registry = gsettings_registry.get_registry()
        for vtype in speechserver.VoiceType:
            registry.set_runtime_value(self._VOICE_SCHEMA, key, value, voice_type=vtype)

    @dbus_service.setter
    def set_rate(self, value: UInt32) -> bool:
        """Sets the current speech rate (0-100, default: 50)."""

        if not isinstance(value, (int, float)):
            return False

        registry = gsettings_registry.get_registry()
        registry.set_runtime_value(self._VOICE_SCHEMA, self.KEY_RATE, value)
        self._sync_runtime_value_to_all_voice_types(self.KEY_RATE, value)

        msg = f"SPEECH MANAGER: Set rate to: {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @dbus_service.command
    def decrease_rate(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Decreases the speech rate."""

        tokens = [
            "SPEECH MANAGER: decrease_rate. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        server.decrease_speech_rate()
        new_rate = max(0, self.get_rate() - 5)
        self.set_rate(new_rate)
        if notify_user and script is not None:
            full = f"{messages.SPEECH_SLOWER} {new_rate}"
            presentation_manager.get_manager().present_message(full, str(new_rate))

        return True

    @dbus_service.command
    def increase_rate(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Increases the speech rate."""

        tokens = [
            "SPEECH MANAGER: increase_rate. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        server.increase_speech_rate()
        new_rate = min(100, self.get_rate() + 5)
        self.set_rate(new_rate)
        if notify_user and script is not None:
            full = f"{messages.SPEECH_FASTER} {new_rate}"
            presentation_manager.get_manager().present_message(full, str(new_rate))

        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_PITCH,
        schema="voice",
        gtype="d",
        default=5.0,
        summary="Speech pitch (0.0-10.0)",
    )
    @dbus_service.getter
    def get_pitch(self) -> float:
        """Returns the current speech pitch."""

        return gsettings_registry.get_registry().layered_lookup(
            self._VOICE_SCHEMA,
            self.KEY_PITCH,
            "d",
            default=5.0,
        )

    @dbus_service.setter
    def set_pitch(self, value: float) -> bool:
        """Sets the current speech pitch (0.0-10.0, default: 5.0)."""

        if not isinstance(value, (int, float)):
            return False

        registry = gsettings_registry.get_registry()
        registry.set_runtime_value(self._VOICE_SCHEMA, self.KEY_PITCH, value)
        self._sync_runtime_value_to_all_voice_types(self.KEY_PITCH, value)

        msg = f"SPEECH MANAGER: Set pitch to: {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @dbus_service.command
    def decrease_pitch(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Decreases the speech pitch"""

        tokens = [
            "SPEECH MANAGER: decrease_pitch. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        server.decrease_speech_pitch()
        new_pitch = max(0.0, self.get_pitch() - 0.5)
        self.set_pitch(new_pitch)
        if notify_user and script is not None:
            full = f"{messages.SPEECH_LOWER} {new_pitch:g}"
            presentation_manager.get_manager().present_message(full, f"{new_pitch:g}")

        return True

    @dbus_service.command
    def increase_pitch(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Increase the speech pitch"""

        tokens = [
            "SPEECH MANAGER: increase_pitch. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        server.increase_speech_pitch()
        new_pitch = min(10.0, self.get_pitch() + 0.5)
        self.set_pitch(new_pitch)
        if notify_user and script is not None:
            full = f"{messages.SPEECH_HIGHER} {new_pitch:g}"
            presentation_manager.get_manager().present_message(full, f"{new_pitch:g}")

        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_PITCH_RANGE,
        schema="voice",
        gtype="d",
        default=5.0,
        summary="Speech inflection / pitch range (0.0-10.0)",
    )
    @dbus_service.getter
    def get_pitch_range(self) -> float:
        """Returns the current speech inflection (pitch range)."""

        return gsettings_registry.get_registry().layered_lookup(
            self._VOICE_SCHEMA,
            self.KEY_PITCH_RANGE,
            "d",
            default=5.0,
        )

    @dbus_service.setter
    def set_pitch_range(self, value: float) -> bool:
        """Sets the current speech inflection / pitch range (0.0-10.0, default: 5.0)."""

        if not isinstance(value, (int, float)):
            return False

        registry = gsettings_registry.get_registry()
        registry.set_runtime_value(self._VOICE_SCHEMA, self.KEY_PITCH_RANGE, value)
        self._sync_runtime_value_to_all_voice_types(self.KEY_PITCH_RANGE, value)

        msg = f"SPEECH MANAGER: Set pitch range to: {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @dbus_service.command
    def decrease_pitch_range(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Decreases the speech inflection (pitch range)."""

        tokens = [
            "SPEECH MANAGER: decrease_pitch_range. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        server.decrease_speech_inflection()
        new_pitch_range = max(0.0, self.get_pitch_range() - 0.5)
        self.set_pitch_range(new_pitch_range)
        if notify_user and script is not None:
            full = f"{messages.SPEECH_LESS_INFLECTION} {new_pitch_range:g}"
            presentation_manager.get_manager().present_message(full, f"{new_pitch_range:g}")

        return True

    @dbus_service.command
    def increase_pitch_range(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Increases the speech inflection (pitch range)."""

        tokens = [
            "SPEECH MANAGER: increase_pitch_range. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        server.increase_speech_inflection()
        new_pitch_range = min(10.0, self.get_pitch_range() + 0.5)
        self.set_pitch_range(new_pitch_range)
        if notify_user and script is not None:
            full = f"{messages.SPEECH_MORE_INFLECTION} {new_pitch_range:g}"
            presentation_manager.get_manager().present_message(full, f"{new_pitch_range:g}")

        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_VOLUME,
        schema="voice",
        gtype="d",
        default=10.0,
        summary="Speech volume (0.0-10.0)",
    )
    @dbus_service.getter
    def get_volume(self) -> float:
        """Returns the current speech volume."""

        return gsettings_registry.get_registry().layered_lookup(
            self._VOICE_SCHEMA,
            self.KEY_VOLUME,
            "d",
            default=10.0,
        )

    @dbus_service.setter
    def set_volume(self, value: float) -> bool:
        """Sets the current speech volume (0.0-10.0, default: 10.0)."""

        if not isinstance(value, (int, float)):
            return False

        registry = gsettings_registry.get_registry()
        registry.set_runtime_value(self._VOICE_SCHEMA, self.KEY_VOLUME, value)
        self._sync_runtime_value_to_all_voice_types(self.KEY_VOLUME, value)

        msg = f"SPEECH MANAGER: Set volume to: {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @dbus_service.command
    def decrease_volume(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Decreases the speech volume"""

        tokens = [
            "SPEECH MANAGER: decrease_volume. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        server.decrease_speech_volume()
        new_volume = max(0.0, self.get_volume() - 0.5)
        self.set_volume(new_volume)
        if notify_user and script is not None:
            full = f"{messages.SPEECH_SOFTER} {new_volume:g}"
            presentation_manager.get_manager().present_message(full, f"{new_volume:g}")

        return True

    @dbus_service.command
    def increase_volume(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Increases the speech volume"""

        tokens = [
            "SPEECH MANAGER: increase_volume. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        server.increase_speech_volume()
        new_volume = min(10.0, self.get_volume() + 0.5)
        self.set_volume(new_volume)
        if notify_user and script is not None:
            full = f"{messages.SPEECH_LOUDER} {new_volume:g}"
            presentation_manager.get_manager().present_message(full, f"{new_volume:g}")

        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_CAPITALIZATION_STYLE,
        schema="speech",
        genum="org.gnome.Orca.CapitalizationStyle",
        default="none",
        summary="Capitalization style (none, spell, icon)",
        migration_key="capitalizationStyle",
    )
    @dbus_service.getter
    def get_capitalization_style(self, app_name: str | None = None) -> str:
        """Returns the current capitalization style."""

        return self._get_setting(
            self.KEY_CAPITALIZATION_STYLE,
            "",
            default="none",
            app_name=app_name,
            genum="org.gnome.Orca.CapitalizationStyle",
        )

    @dbus_service.setter
    def set_capitalization_style(self, value: str) -> bool:
        """Sets the capitalization style."""

        try:
            style = CapitalizationStyle[value.upper()]
        except KeyError:
            msg = f"SPEECH MANAGER: Invalid capitalization style: {value}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        msg = f"SPEECH MANAGER: Setting capitalization style to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA,
            self.KEY_CAPITALIZATION_STYLE,
            style.string_name,
        )
        self.update_capitalization_style()
        return True

    @dbus_service.command
    def cycle_capitalization_style(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Cycle through the speech-dispatcher capitalization styles."""

        tokens = [
            "SPEECH MANAGER: cycle_capitalization_style. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        current_style = self.get_capitalization_style()
        if current_style == "none":
            self.set_capitalization_style("spell")
            full = messages.CAPITALIZATION_SPELL_FULL
            brief = messages.CAPITALIZATION_SPELL_BRIEF
        elif current_style == "spell":
            self.set_capitalization_style("icon")
            full = messages.CAPITALIZATION_ICON_FULL
            brief = messages.CAPITALIZATION_ICON_BRIEF
        else:
            self.set_capitalization_style("none")
            full = messages.CAPITALIZATION_NONE_FULL
            brief = messages.CAPITALIZATION_NONE_BRIEF

        if script is not None and notify_user:
            presentation_manager.get_manager().present_message(full, brief)
        return True

    def update_capitalization_style(self) -> bool:
        """Updates the capitalization style on the speech server."""

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        server.update_capitalization_style(self.get_capitalization_style())
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_PUNCTUATION_LEVEL,
        schema="speech",
        genum="org.gnome.Orca.PunctuationStyle",
        default="most",
        summary="Punctuation verbosity level (none, some, most, all)",
        migration_key="verbalizePunctuationStyle",
    )
    @dbus_service.getter
    def get_punctuation_level(self, app_name: str | None = None) -> str:
        """Returns the current punctuation level."""

        return self._get_setting(
            self.KEY_PUNCTUATION_LEVEL,
            "",
            default="most",
            app_name=app_name,
            genum="org.gnome.Orca.PunctuationStyle",
        )

    @dbus_service.setter
    def set_punctuation_level(self, value: str) -> bool:
        """Sets the punctuation level."""

        try:
            style = PunctuationStyle[value.upper()]
        except KeyError:
            msg = f"SPEECH MANAGER: Invalid punctuation level: {value}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        msg = f"SPEECH MANAGER: Setting punctuation level to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA,
            self.KEY_PUNCTUATION_LEVEL,
            style.string_name,
        )
        self.update_punctuation_level()
        return True

    @dbus_service.command
    def cycle_punctuation_level(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Cycles through punctuation levels for speech."""

        tokens = [
            "SPEECH MANAGER: cycle_punctuation_level. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        current_level = self.get_punctuation_level()
        if current_level == "none":
            self.set_punctuation_level("some")
            full = messages.PUNCTUATION_SOME_FULL
            brief = messages.PUNCTUATION_SOME_BRIEF
        elif current_level == "some":
            self.set_punctuation_level("most")
            full = messages.PUNCTUATION_MOST_FULL
            brief = messages.PUNCTUATION_MOST_BRIEF
        elif current_level == "most":
            self.set_punctuation_level("all")
            full = messages.PUNCTUATION_ALL_FULL
            brief = messages.PUNCTUATION_ALL_BRIEF
        else:
            self.set_punctuation_level("none")
            full = messages.PUNCTUATION_NONE_FULL
            brief = messages.PUNCTUATION_NONE_BRIEF

        if script is not None and notify_user:
            presentation_manager.get_manager().present_message(full, brief)
        return True

    def update_punctuation_level(self) -> bool:
        """Updates the punctuation level on the speech server."""

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        level_str = self.get_punctuation_level()
        server.update_punctuation_level(PunctuationStyle[level_str.upper()])
        return True

    def update_synthesizer(self, server_id: str | None = "") -> None:
        """Updates the synthesizer to the specified id or value from settings."""

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        active_id = server.get_output_module()
        if not server_id:
            server_id = self._get_setting(self.KEY_SYNTHESIZER, "s", "")

        if server_id and server_id != active_id:
            msg = f"SPEECH MANAGER: Updating synthesizer from {active_id} to {server_id}."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            server.set_output_module(server_id)

    @dbus_service.command
    def cycle_synthesizer(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Cycles through available speech synthesizers."""

        tokens = [
            "SPEECH MANAGER: cycle_synthesizer. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        server = self._get_server()
        if server is None:
            msg = "SPEECH MANAGER: Cannot get speech server."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        available = server.list_output_modules()
        if not available:
            msg = "SPEECH MANAGER: Cannot get output modules."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        current = server.get_output_module()
        if not current:
            msg = "SPEECH MANAGER: Cannot get current output module."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        try:
            index = available.index(current) + 1
            if index == len(available):
                index = 0
        except ValueError:
            index = 0

        server.set_output_module(available[index])
        if script is not None and notify_user:
            presentation_manager.get_manager().present_message(available[index])
        return True

    def get_speech_is_enabled_and_not_muted(self) -> bool:
        """Returns whether speech is enabled and not muted."""

        return self.get_speech_is_enabled() and not self.get_speech_is_muted()

    @dbus_service.getter
    def get_speech_is_muted(self) -> bool:
        """Returns whether speech output is temporarily muted."""

        return self._mute_speech

    @dbus_service.setter
    def set_speech_is_muted(self, value: bool) -> bool:
        """Sets whether speech output is temporarily muted."""

        msg = f"SPEECH MANAGER: Setting speech muted to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._mute_speech = value
        return True

    @dbus_service.getter
    def get_available_voice_sets(self) -> list[str]:
        """Returns the valid values for the active voice set."""

        result = [gsettings_registry.PRIMARY_VOICE_SET, *sorted(self.get_voice_set_names())]
        msg = f"SPEECH MANAGER: Available voice sets: {result}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    @dbus_service.getter
    def get_active_voice_set(self) -> str:
        """Returns the name of the active voice set used for speech output."""

        return self._active_voice_set

    @dbus_service.setter
    def set_active_voice_set(self, name: str) -> bool:
        """Sets the active voice set used for speech output."""

        if name != gsettings_registry.PRIMARY_VOICE_SET and name not in self.get_voice_set_names():
            msg = f"SPEECH MANAGER: Ignoring unknown voice set {name!r}."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False
        msg = f"SPEECH MANAGER: Setting active voice set to {name}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._active_voice_set = name
        return True

    def _voice_set_display_name(self, set_id: str) -> str:
        """Returns the user-facing name for the given voice set."""

        if set_id == gsettings_registry.PRIMARY_VOICE_SET:
            return guilabels.VOICE_SET_GLOBAL
        return language_utilities.get_language_display_name(set_id, in_own_language=True)

    @dbus_service.parameterized_command
    def activate_voice_set(
        self,
        set_id: str,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Makes set_id the active voice set and announces the change."""

        tokens = ["SPEECH MANAGER: activate_voice_set", set_id, "Script:", script, "Event:", event]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not self.set_active_voice_set(set_id):
            return True

        if script is not None and notify_user:
            presentation_manager.get_manager().present_message(
                self._voice_set_display_name(set_id), voice_type=speechserver.VoiceType.DEFAULT
            )
        return True

    @dbus_service.command
    def cycle_voice_set(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Switches to the next available voice set, wrapping after the last."""

        tokens = ["SPEECH MANAGER: cycle_voice_set. Script:", script, "Event:", event]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        available = self.get_available_voice_sets()
        try:
            index = available.index(self._active_voice_set) + 1
        except ValueError:
            index = 0
        if index == len(available):
            index = 0

        return self.activate_voice_set(available[index], script, event, notify_user)

    @gsettings_registry.get_registry().gsetting(
        key=KEY_ENABLE,
        schema="speech",
        gtype="b",
        default=True,
        summary="Enable speech output",
        migration_key="enableSpeech",
    )
    @dbus_service.getter
    def get_speech_is_enabled(self) -> bool:
        """Returns whether the speech server is enabled. See also is-muted."""

        return self._get_setting(self.KEY_ENABLE, "b", True)

    @dbus_service.setter
    def set_speech_is_enabled(self, value: bool) -> bool:
        """Sets whether the speech server is enabled. See also is-muted."""

        if value == self.get_speech_is_enabled():
            return True

        msg = f"SPEECH MANAGER: Setting speech enabled to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)

        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA, self.KEY_ENABLE, value
        )
        if value:
            self.start_speech()
            presentation_manager.get_manager().present_message(messages.SPEECH_ENABLED)
            systemd.get_manager().set_status("Speech", "enabled")
        else:
            presentation_manager.get_manager().present_message(messages.SPEECH_DISABLED)
            self.shutdown_speech()
            systemd.get_manager().set_status("Speech", "disabled")

        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SPEAK_NUMBERS_AS_DIGITS,
        schema="speech",
        gtype="b",
        default=False,
        summary="Speak numbers as individual digits",
        migration_key="speakNumbersAsDigits",
    )
    @dbus_service.getter
    def get_speak_numbers_as_digits(self, app_name: str | None = None) -> bool:
        """Returns whether numbers are spoken as digits."""

        return self._get_setting(self.KEY_SPEAK_NUMBERS_AS_DIGITS, "b", False, app_name=app_name)

    @dbus_service.setter
    def set_speak_numbers_as_digits(self, value: bool) -> bool:
        """Sets whether numbers are spoken as digits."""

        msg = f"SPEECH MANAGER: Setting speak numbers as digits to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA,
            self.KEY_SPEAK_NUMBERS_AS_DIGITS,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_USE_COLOR_NAMES,
        schema="speech",
        gtype="b",
        default=True,
        summary="Use color names instead of values",
        migration_key="useColorNames",
    )
    @dbus_service.getter
    def get_use_color_names(self, app_name: str | None = None) -> bool:
        """Returns whether colors are announced by name or as RGB values."""

        return self._get_setting(self.KEY_USE_COLOR_NAMES, "b", True, app_name=app_name)

    @dbus_service.setter
    def set_use_color_names(self, value: bool) -> bool:
        """Sets whether colors are announced by name or as RGB values."""

        msg = f"SPEECH MANAGER: Setting use color names to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA,
            self.KEY_USE_COLOR_NAMES,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_INSERT_PAUSES_BETWEEN_UTTERANCES,
        schema="speech",
        gtype="b",
        default=True,
        summary="Insert pauses between utterances",
        migration_key="enablePauseBreaks",
    )
    @dbus_service.getter
    def get_insert_pauses_between_utterances(self, app_name: str | None = None) -> bool:
        """Returns whether pauses are inserted between utterances, e.g. between name and role."""

        return self._get_setting(
            self.KEY_INSERT_PAUSES_BETWEEN_UTTERANCES, "b", True, app_name=app_name
        )

    @dbus_service.setter
    def set_insert_pauses_between_utterances(self, value: bool) -> bool:
        """Sets whether pauses are inserted between utterances, e.g. between name and role."""

        msg = f"SPEECH MANAGER: Setting insert pauses to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA,
            self.KEY_INSERT_PAUSES_BETWEEN_UTTERANCES,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_USE_PRONUNCIATION_DICTIONARY,
        schema="speech",
        gtype="b",
        default=True,
        summary="Apply user pronunciation dictionary",
        migration_key="usePronunciationDictionary",
    )
    @dbus_service.getter
    def get_use_pronunciation_dictionary(self, app_name: str | None = None) -> bool:
        """Returns whether the user's pronunciation dictionary should be applied."""

        return self._get_setting(
            self.KEY_USE_PRONUNCIATION_DICTIONARY, "b", True, app_name=app_name
        )

    @dbus_service.setter
    def set_use_pronunciation_dictionary(self, value: bool) -> bool:
        """Sets whether the user's pronunciation dictionary should be applied."""

        msg = f"SPEECH MANAGER: Setting use pronunciation dictionary to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA,
            self.KEY_USE_PRONUNCIATION_DICTIONARY,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_AUTO_LANGUAGE_SWITCHING,
        schema="speech",
        gtype="b",
        default=True,
        summary="Automatically switch voice based on document content language",
        migration_key="enableAutoLanguageSwitching",
    )
    @dbus_service.getter
    def get_auto_language_switching(self, app_name: str | None = None) -> bool:
        """Returns whether automatic language switching for document content is enabled."""

        return self._get_setting(self.KEY_AUTO_LANGUAGE_SWITCHING, "b", True, app_name=app_name)

    @dbus_service.setter
    def set_auto_language_switching(self, value: bool) -> bool:
        """Sets whether automatic language switching for document content is enabled."""

        msg = f"SPEECH MANAGER: Setting auto language switching for content to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA,
            self.KEY_AUTO_LANGUAGE_SWITCHING,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_AUTO_LANGUAGE_SWITCHING_UI,
        schema="speech",
        gtype="b",
        default=False,
        summary="Automatically switch voice based on UI element language",
    )
    @dbus_service.getter
    def get_auto_language_switching_ui(self, app_name: str | None = None) -> bool:
        """Returns whether automatic language switching for UI elements is enabled."""

        return self._get_setting(self.KEY_AUTO_LANGUAGE_SWITCHING_UI, "b", False, app_name=app_name)

    @dbus_service.setter
    def set_auto_language_switching_ui(self, value: bool) -> bool:
        """Sets whether automatic language switching for UI elements is enabled."""

        msg = f"SPEECH MANAGER: Setting auto language switching for UI to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA,
            self.KEY_AUTO_LANGUAGE_SWITCHING_UI,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_ONLY_SWITCH_CONFIGURED_LANGUAGES,
        schema="speech",
        gtype="b",
        default=False,
        summary="Only switch languages that have a configured voice set",
    )
    @dbus_service.getter
    def get_only_switch_configured_languages(self, app_name: str | None = None) -> bool:
        """Returns whether language switching is limited to configured voice sets."""

        return self._get_setting(
            self.KEY_ONLY_SWITCH_CONFIGURED_LANGUAGES, "b", False, app_name=app_name
        )

    @dbus_service.setter
    def set_only_switch_configured_languages(self, value: bool) -> bool:
        """Sets whether language switching is limited to configured voice sets."""

        msg = f"SPEECH MANAGER: Setting only switch configured languages to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self.SPEECH_SCHEMA,
            self.KEY_ONLY_SWITCH_CONFIGURED_LANGUAGES,
            value,
        )
        return True

    @dbus_service.command
    def toggle_speech(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Toggles speech on and off."""

        tokens = [
            "SPEECH MANAGER: toggle_speech. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if script is not None:
            presentation_manager.get_manager().interrupt_presentation()
        if self.get_speech_is_muted():
            self.set_speech_is_muted(False)
            if script is not None and notify_user:
                presentation_manager.get_manager().present_message(messages.SPEECH_ENABLED)
        elif not self.get_speech_is_enabled():
            gsettings_registry.get_registry().set_runtime_value(
                self.SPEECH_SCHEMA, self.KEY_ENABLE, True
            )
            self._init_server()
            if script is not None and notify_user:
                presentation_manager.get_manager().present_message(messages.SPEECH_ENABLED)
        else:
            if script is not None and notify_user:
                presentation_manager.get_manager().present_message(messages.SPEECH_DISABLED)
            gsettings_registry.get_registry().remove_runtime_value(
                self.SPEECH_SCHEMA, self.KEY_ENABLE
            )
            if not self.get_speech_is_enabled():
                self.shutdown_speech()
            else:
                self.set_speech_is_muted(True)
        return True

    def create_voices_preferences_grid(self, app_name: str = "") -> VoicesPreferencesGrid:
        """Returns the GtkGrid containing the voices preferences UI."""

        # pylint: disable-next=import-outside-toplevel
        from .speech_manager_preferences_grid import VoicesPreferencesGrid

        return VoicesPreferencesGrid(self, app_name=app_name)

    def create_voice_types_preferences_grid(
        self, voices_grid: VoicesPreferencesGrid
    ) -> VoiceTypesPreferencesGrid:
        """Returns the GtkGrid containing the voice types preferences UI."""

        # pylint: disable-next=import-outside-toplevel
        from .speech_manager_preferences_grid import VoiceTypesPreferencesGrid

        return VoiceTypesPreferencesGrid(voices_grid)


_manager: SpeechManager = SpeechManager()


def get_manager() -> SpeechManager:
    """Returns the Speech Manager"""

    return _manager



# Orca
#
# Copyright 2024 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for bypass mode."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .bypass_mode_manager import BypassModeManager


def get_commands(owner: BypassModeManager) -> list[Command]:
    """Returns commands for bypass mode."""

    kb_alt_backspace = keybindings.KeyBinding("BackSpace", keybindings.ALT_MODIFIER_MASK)
    return [
        KeyboardCommand(
            owner.COMMAND_NAME,
            owner.toggle_enabled,
            owner.GROUP_LABEL,
            cmdnames.BYPASS_MODE_TOGGLE,
            desktop_keybinding=kb_alt_backspace,
            laptop_keybinding=kb_alt_backspace,
        ),
    ]



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2011-2023 Igalia, S.L.
# Copyright 2023 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for table navigation."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .table_navigator import TableNavigator


def get_commands(owner: TableNavigator) -> list[Command]:  # pylint: disable=too-many-locals
    """Returns commands for table navigation."""

    kb_orca_shift_t = keybindings.KeyBinding("t", keybindings.ORCA_SHIFT_MODIFIER_MASK)
    kb_shift_alt_left = keybindings.KeyBinding("Left", keybindings.SHIFT_ALT_MODIFIER_MASK)
    kb_shift_alt_right = keybindings.KeyBinding("Right", keybindings.SHIFT_ALT_MODIFIER_MASK)
    kb_shift_alt_up = keybindings.KeyBinding("Up", keybindings.SHIFT_ALT_MODIFIER_MASK)
    kb_shift_alt_down = keybindings.KeyBinding("Down", keybindings.SHIFT_ALT_MODIFIER_MASK)
    kb_shift_alt_home = keybindings.KeyBinding("Home", keybindings.SHIFT_ALT_MODIFIER_MASK)
    kb_shift_alt_end = keybindings.KeyBinding("End", keybindings.SHIFT_ALT_MODIFIER_MASK)
    kb_orca_alt_shift_left = keybindings.KeyBinding(
        "Left",
        keybindings.ORCA_ALT_SHIFT_MODIFIER_MASK,
    )
    kb_orca_alt_shift_right = keybindings.KeyBinding(
        "Right",
        keybindings.ORCA_ALT_SHIFT_MODIFIER_MASK,
    )
    kb_orca_alt_shift_up = keybindings.KeyBinding(
        "Up",
        keybindings.ORCA_ALT_SHIFT_MODIFIER_MASK,
    )
    kb_orca_alt_shift_down = keybindings.KeyBinding(
        "Down",
        keybindings.ORCA_ALT_SHIFT_MODIFIER_MASK,
    )
    kb_orca_shift_r = keybindings.KeyBinding("r", keybindings.ORCA_SHIFT_MODIFIER_MASK)
    kb_orca_shift_r_2 = keybindings.KeyBinding(
        "r",
        keybindings.ORCA_SHIFT_MODIFIER_MASK,
        click_count=2,
    )
    kb_orca_shift_c = keybindings.KeyBinding("c", keybindings.ORCA_SHIFT_MODIFIER_MASK)
    kb_orca_shift_c_2 = keybindings.KeyBinding(
        "c",
        keybindings.ORCA_SHIFT_MODIFIER_MASK,
        click_count=2,
    )

    return [
        KeyboardCommand(
            "table_navigator_toggle_enabled",
            owner.toggle_enabled,
            owner.GROUP_LABEL,
            cmdnames.TABLE_NAVIGATION_TOGGLE,
            desktop_keybinding=kb_orca_shift_t,
            laptop_keybinding=kb_orca_shift_t,
            is_group_toggle=True,
        ),
        KeyboardCommand(
            "table_cell_left",
            owner.move_left,
            owner.GROUP_LABEL,
            cmdnames.TABLE_CELL_LEFT,
            desktop_keybinding=kb_shift_alt_left,
            laptop_keybinding=kb_shift_alt_left,
        ),
        KeyboardCommand(
            "table_cell_right",
            owner.move_right,
            owner.GROUP_LABEL,
            cmdnames.TABLE_CELL_RIGHT,
            desktop_keybinding=kb_shift_alt_right,
            laptop_keybinding=kb_shift_alt_right,
        ),
        KeyboardCommand(
            "table_cell_up",
            owner.move_up,
            owner.GROUP_LABEL,
            cmdnames.TABLE_CELL_UP,
            desktop_keybinding=kb_shift_alt_up,
            laptop_keybinding=kb_shift_alt_up,
        ),
        KeyboardCommand(
            "table_cell_down",
            owner.move_down,
            owner.GROUP_LABEL,
            cmdnames.TABLE_CELL_DOWN,
            desktop_keybinding=kb_shift_alt_down,
            laptop_keybinding=kb_shift_alt_down,
        ),
        KeyboardCommand(
            "table_cell_first",
            owner.move_to_first_cell,
            owner.GROUP_LABEL,
            cmdnames.TABLE_CELL_FIRST,
            desktop_keybinding=kb_shift_alt_home,
            laptop_keybinding=kb_shift_alt_home,
        ),
        KeyboardCommand(
            "table_cell_last",
            owner.move_to_last_cell,
            owner.GROUP_LABEL,
            cmdnames.TABLE_CELL_LAST,
            desktop_keybinding=kb_shift_alt_end,
            laptop_keybinding=kb_shift_alt_end,
        ),
        KeyboardCommand(
            "table_cell_beginning_of_row",
            owner.move_to_beginning_of_row,
            owner.GROUP_LABEL,
            cmdnames.TABLE_CELL_BEGINNING_OF_ROW,
            desktop_keybinding=kb_orca_alt_shift_left,
            laptop_keybinding=kb_orca_alt_shift_left,
        ),
        KeyboardCommand(
            "table_cell_end_of_row",
            owner.move_to_end_of_row,
            owner.GROUP_LABEL,
            cmdnames.TABLE_CELL_END_OF_ROW,
            desktop_keybinding=kb_orca_alt_shift_right,
            laptop_keybinding=kb_orca_alt_shift_right,
        ),
        KeyboardCommand(
            "table_cell_top_of_column",
            owner.move_to_top_of_column,
            owner.GROUP_LABEL,
            cmdnames.TABLE_CELL_TOP_OF_COLUMN,
            desktop_keybinding=kb_orca_alt_shift_up,
            laptop_keybinding=kb_orca_alt_shift_up,
        ),
        KeyboardCommand(
            "table_cell_bottom_of_column",
            owner.move_to_bottom_of_column,
            owner.GROUP_LABEL,
            cmdnames.TABLE_CELL_BOTTOM_OF_COLUMN,
            desktop_keybinding=kb_orca_alt_shift_down,
            laptop_keybinding=kb_orca_alt_shift_down,
        ),
        KeyboardCommand(
            "set_dynamic_column_headers_row",
            owner.set_dynamic_column_headers_row,
            owner.GROUP_LABEL,
            cmdnames.DYNAMIC_COLUMN_HEADER_SET,
            desktop_keybinding=kb_orca_shift_r,
            laptop_keybinding=kb_orca_shift_r,
        ),
        KeyboardCommand(
            "clear_dynamic_column_headers_row",
            owner.clear_dynamic_column_headers_row,
            owner.GROUP_LABEL,
            cmdnames.DYNAMIC_COLUMN_HEADER_CLEAR,
            desktop_keybinding=kb_orca_shift_r_2,
            laptop_keybinding=kb_orca_shift_r_2,
        ),
        KeyboardCommand(
            "set_dynamic_row_headers_column",
            owner.set_dynamic_row_headers_column,
            owner.GROUP_LABEL,
            cmdnames.DYNAMIC_ROW_HEADER_SET,
            desktop_keybinding=kb_orca_shift_c,
            laptop_keybinding=kb_orca_shift_c,
        ),
        KeyboardCommand(
            "clear_dynamic_row_headers_column",
            owner.clear_dynamic_row_headers_column,
            owner.GROUP_LABEL,
            cmdnames.DYNAMIC_ROW_HEADER_CLEAR,
            desktop_keybinding=kb_orca_shift_c_2,
            laptop_keybinding=kb_orca_shift_c_2,
        ),
    ]



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2016-2026 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-locals

"""Module for presenting system information"""

from __future__ import annotations

import time
from enum import Enum
from typing import TYPE_CHECKING, Any

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

_PSUTIL_AVAILABLE = False  # pylint: disable=invalid-name
try:
    import psutil  # type: ignore

    _PSUTIL_AVAILABLE = True  # pylint: disable=invalid-name
except ModuleNotFoundError:
    pass

from . import (
    ax_device_manager,
    dbus_service,
    debug,
    gsettings_registry,
    guilabels,
    input_event,
    keynames,
    messages,
    presentation_manager,
    system_information_presenter_command_definitions,
)
from .extension import Extension

if TYPE_CHECKING:
    from .command import Command


class DateFormat(Enum):
    """Date format enumeration with format strings."""

    LOCALE = "%x"
    NUMBERS_DM = "%d/%m"
    NUMBERS_MD = "%m/%d"
    NUMBERS_DMY = "%d/%m/%Y"
    NUMBERS_MDY = "%m/%d/%Y"
    NUMBERS_YMD = "%Y/%m/%d"
    FULL_DM = "%A, %-d %B"
    FULL_MD = "%A, %B %-d"
    FULL_DMY = "%A, %-d %B, %Y"
    FULL_MDY = "%A, %B %-d, %Y"
    FULL_YMD = "%Y. %B %-d, %A"
    ABBREVIATED_DM = "%a, %-d %b"
    ABBREVIATED_MD = "%a, %b %-d"
    ABBREVIATED_DMY = "%a, %-d %b, %Y"
    ABBREVIATED_MDY = "%a, %b %-d, %Y"
    ABBREVIATED_YMD = "%Y. %b %-d, %a"

    @property
    def string_name(self) -> str:
        """Returns the lowercase string name for this enum value."""

        return self.name.lower()


class TimeFormat(Enum):
    """Time format enumeration with format strings."""

    LOCALE = "%X"
    TWELVE_HM = "%I:%M %p"
    TWELVE_HMS = "%I:%M:%S %p"
    TWENTYFOUR_HM = "%H:%M"
    TWENTYFOUR_HMS = "%H:%M:%S"
    TWENTYFOUR_HM_WITH_WORDS = messages.TIME_FORMAT_24_HM_WITH_WORDS
    TWENTYFOUR_HMS_WITH_WORDS = messages.TIME_FORMAT_24_HMS_WITH_WORDS

    @property
    def string_name(self) -> str:
        """Returns the lowercase string name for this enum value."""

        return self.name.lower()


if TYPE_CHECKING:
    from .scripts import default
    from .system_information_presenter_preferences_grid import TimeAndDatePreferencesGrid


@gsettings_registry.get_registry().gsettings_schema(
    "org.gnome.Orca.SystemInformation",
    name="system-information",
)
class SystemInformationPresenter(Extension):
    """Provides commands to present system information."""

    _SCHEMA = "system-information"
    KEY_DATE_FORMAT = "date-format"
    KEY_TIME_FORMAT = "time-format"

    def _get_setting(self, key: str, gtype: str, default: Any) -> Any:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            key,
            gtype,
            default=default,
        )

    GROUP_LABEL = guilabels.KB_GROUP_SYSTEM_INFORMATION

    def _get_commands(self) -> list[Command]:
        return system_information_presenter_command_definitions.get_commands(self)

    def create_time_and_date_preferences_grid(self) -> TimeAndDatePreferencesGrid:
        """Returns the GtkGrid containing the time and date preferences UI."""

        # pylint: disable-next=import-outside-toplevel
        from .system_information_presenter_preferences_grid import TimeAndDatePreferencesGrid

        return TimeAndDatePreferencesGrid(self)

    def _get_date_format_string(self) -> str:
        """Returns the current date format string for internal use."""

        return self._get_setting(self.KEY_DATE_FORMAT, "s", "%x")

    def _get_time_format_string(self) -> str:
        """Returns the current time format string for internal use."""

        return self._get_setting(self.KEY_TIME_FORMAT, "s", "%X")

    def _set_date_format_string(self, value: str) -> bool:
        """Sets the date format string directly for internal use."""

        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_DATE_FORMAT, value
        )
        return True

    def _set_time_format_string(self, value: str) -> bool:
        """Sets the time format string directly for internal use."""

        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_TIME_FORMAT, value
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_DATE_FORMAT,
        schema="system-information",
        gtype="s",
        default="%x",
        summary="Date format string",
        migration_key="presentDateFormat",
    )
    @dbus_service.getter
    def get_date_format(self) -> str:
        """Returns the current date format name."""

        string_value = self._get_date_format_string()
        for fmt in DateFormat:
            if fmt.value == string_value:
                return fmt.string_name
        return string_value

    @dbus_service.setter
    def set_date_format(self, value: str) -> bool:
        """Sets the date format."""

        try:
            fmt = DateFormat[value.upper()]
        except KeyError:
            msg = f"SYSTEM INFORMATION PRESENTER: Invalid date format: {value}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        msg = f"SYSTEM INFORMATION PRESENTER: Setting date format to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_DATE_FORMAT, fmt.value
        )
        return True

    @dbus_service.getter
    def get_available_date_formats(self) -> list[str]:
        """Returns a list of available date format names."""

        return [fmt.string_name for fmt in DateFormat]

    @gsettings_registry.get_registry().gsetting(
        key=KEY_TIME_FORMAT,
        schema="system-information",
        gtype="s",
        default="%X",
        summary="Time format string",
        migration_key="presentTimeFormat",
    )
    @dbus_service.getter
    def get_time_format(self) -> str:
        """Returns the current time format name."""

        string_value = self._get_time_format_string()
        for fmt in TimeFormat:
            if fmt.value == string_value:
                return fmt.string_name
        return string_value

    @dbus_service.setter
    def set_time_format(self, value: str) -> bool:
        """Sets the time format."""

        try:
            fmt = TimeFormat[value.upper()]
        except KeyError:
            msg = f"SYSTEM INFORMATION PRESENTER: Invalid time format: {value}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        msg = f"SYSTEM INFORMATION PRESENTER: Setting time format to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_TIME_FORMAT, fmt.value
        )
        return True

    @dbus_service.getter
    def get_available_time_formats(self) -> list[str]:
        """Returns a list of available time format names."""

        return [fmt.string_name for fmt in TimeFormat]

    @dbus_service.command
    def present_time(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current time."""

        tokens = [
            "SYSTEM INFORMATION PRESENTER: present_time. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        time_format = self._get_time_format_string()
        presentation_manager.get_manager().present_message(
            time.strftime(time_format, time.localtime()),
        )
        return True

    @dbus_service.command
    def present_date(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current date."""

        tokens = [
            "SYSTEM INFORMATION PRESENTER: present_date. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        date_format = self._get_date_format_string()
        presentation_manager.get_manager().present_message(
            time.strftime(date_format, time.localtime()),
        )
        return True

    @dbus_service.command
    def present_battery_status(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the battery status."""

        tokens = [
            "SYSTEM INFORMATION PRESENTER: present_battery_status. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        battery = psutil.sensors_battery() if _PSUTIL_AVAILABLE else None
        if battery is None:
            presentation_manager.get_manager().present_message(messages.BATTERY_STATUS_UNKNOWN)
            return True
        if battery.power_plugged:
            msg = f"{messages.BATTERY_LEVEL % battery.percent} {messages.BATTERY_PLUGGED_IN_TRUE}"
        else:
            msg = f"{messages.BATTERY_LEVEL % battery.percent} {messages.BATTERY_PLUGGED_IN_FALSE}"

        presentation_manager.get_manager().present_message(msg)
        return True

    @dbus_service.command
    def present_cpu_and_memory_usage(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the cpu and memory usage."""

        tokens = [
            "SYSTEM INFORMATION PRESENTER: present_cpu_and_memory_usage. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not _PSUTIL_AVAILABLE:
            presentation_manager.get_manager().present_message(
                messages.CPU_AND_MEMORY_USAGE_UNKNOWN,
            )
            return True

        cpu_usage = round(psutil.cpu_percent())

        memory = psutil.virtual_memory()
        memory_percent = round(memory.percent)
        if memory.total > 1024**3:
            details = messages.memory_usage_gb(memory.used / (1024**3), memory.total / (1024**3))
        else:
            details = messages.memory_usage_mb(memory.used / (1024**2), memory.total / (1024**2))

        msg = f"{messages.CPU_AND_MEMORY_USAGE_LEVELS % (cpu_usage, memory_percent)}. {details}"
        presentation_manager.get_manager().present_message(msg)
        return True

    @dbus_service.command
    def present_modifier_keys_state(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the state of locked modifier keys. Requires AT-SPI 2.59.0 or later."""

        tokens = [
            "SYSTEM INFORMATION PRESENTER: present_modifier_keys_state. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        manager = ax_device_manager.get_manager()
        if not manager.is_active():
            presentation_manager.get_manager().present_message(
                messages.MODIFIER_KEYS_STATE_UNAVAILABLE,
            )
            return True

        locked = manager.get_locked_modifiers()
        modifiers = [
            ("Caps_Lock", Atspi.ModifierType.SHIFTLOCK),
            ("Num_Lock", Atspi.ModifierType.NUMLOCK),
        ]

        parts: list[str] = []
        for key_name, modifier_type in modifiers:
            bit = 1 << modifier_type
            is_locked = bool(locked & bit)
            msg = (
                f"SYSTEM INFORMATION PRESENTER: {key_name}: "
                f"modifier_type={modifier_type}, bit={bit:#x}, "
                f"locked & bit={locked & bit:#x}, is_locked={is_locked}"
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)

            localized_name = keynames.get_key_name(key_name) or key_name
            if is_locked:
                state = messages.LOCKING_KEY_STATE_ON
            else:
                state = messages.LOCKING_KEY_STATE_OFF
            parts.append(f"{localized_name} {state}")

        presentation_manager.get_manager().present_message(". ".join(parts))
        return True


_presenter = SystemInformationPresenter()


def get_presenter() -> SystemInformationPresenter:
    """Returns the system-information-presenter singleton."""

    return _presenter



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2016-2026 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for system information."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .system_information_presenter import SystemInformationPresenter


def get_commands(owner: SystemInformationPresenter) -> list[Command]:
    """Returns commands for system information."""

    kb_orca_t = keybindings.KeyBinding("t", keybindings.ORCA_MODIFIER_MASK)
    kb_orca_t_2 = keybindings.KeyBinding("t", keybindings.ORCA_MODIFIER_MASK, click_count=2)

    return [
        KeyboardCommand(
            "presentTimeHandler",
            owner.present_time,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_CURRENT_TIME,
            desktop_keybinding=kb_orca_t,
            laptop_keybinding=kb_orca_t,
        ),
        KeyboardCommand(
            "presentDateHandler",
            owner.present_date,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_CURRENT_DATE,
            desktop_keybinding=kb_orca_t_2,
            laptop_keybinding=kb_orca_t_2,
        ),
        KeyboardCommand(
            "present_battery_status",
            owner.present_battery_status,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_BATTERY_STATUS,
        ),
        KeyboardCommand(
            "present_cpu_and_memory_usage",
            owner.present_cpu_and_memory_usage,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_CPU_AND_MEMORY_USAGE,
        ),
        KeyboardCommand(
            "present_modifier_keys_state",
            owner.present_modifier_keys_state,
            owner.GROUP_LABEL,
            cmdnames.PRESENT_MODIFIER_KEYS_STATE,
        ),
    ]



# Orca
#
# Copyright (C) 2015 Igalia, S.L.
#
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Custom script utilities for Evolution."""

from __future__ import annotations

from typing import TYPE_CHECKING

from orca import focus_manager, input_event_manager
from orca.ax_object import AXObject
from orca.ax_utilities import AXUtilities
from orca.scripts import web

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi


class Utilities(web.Utilities):
    """Custom script utilities for Evolution."""

    def is_message_list_status_cell(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a message list status cell."""

        if not self.is_message_list_toggle_cell(obj):
            return False

        headers = AXUtilities.get_column_headers(obj)
        if not headers:
            return False

        return headers[0] and AXObject.get_name(headers[0]) != AXObject.get_name(obj)

    def is_message_list_toggle_cell(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a message list toggle cell."""

        if self.in_document_content(obj):
            return False

        if not AXObject.get_name(obj):
            return False

        return self.has_meaningful_toggle_action(obj)

    def is_ignorable_event_from_document_preview(self, event: Atspi.Event) -> bool:
        """Returns True if event is from a document preview and can be ignored."""

        if not self.is_document_preview(event.source):
            return False

        if not input_event_manager.get_manager().last_event_was_unmodified_arrow():
            return False

        focus = focus_manager.get_manager().get_locus_of_focus()
        if self.in_document_content(focus):
            return False
        if not AXUtilities.is_table_cell(focus):
            return False
        return AXUtilities.is_tree_or_tree_table_descendant(focus)

    def is_document_preview(self, obj: Atspi.Accessible) -> bool:
        """Returns True if obj is or descends from the preview document."""

        if not self.in_document_content(obj):
            return False

        if AXUtilities.is_document(obj):
            document = obj
        else:
            document = AXUtilities.find_ancestor(obj, AXUtilities.is_document)
        if not document:
            return False

        return bool(AXUtilities.find_ancestor(document, AXUtilities.is_page_tab))



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Provides a graphical speech monitor, mainly for development tasks."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, Gtk

from . import guilabels

if TYPE_CHECKING:
    from collections.abc import Callable

MIN_FONT_SIZE = 8
MAX_FONT_SIZE = 72
ZOOM_STEP = 2


class SpeechMonitor(Gtk.Window):  # pylint: disable=too-many-instance-attributes
    """Displays a GUI speech monitor showing a scrolling log of spoken text."""

    _shared_css_provider: Gtk.CssProvider | None = None

    # pylint: disable-next=too-many-statements
    def __init__(
        self,
        font_size: int = 14,
        foreground: str = "#ffffff",
        background: str = "#000000",
        on_close: Callable[[], None] | None = None,
    ) -> None:
        """Create a new SpeechMonitor."""

        # pylint: disable=no-member

        super().__init__()
        self._on_close = on_close
        self._font_size = font_size
        self._default_font_size = font_size
        self._foreground = foreground
        self._background = background
        self.set_title(guilabels.SPEECH_MONITOR)
        self.set_default_size(1000, 400)
        self.set_icon_name("orca")

        titlebar = Gtk.Box()
        self.set_titlebar(titlebar)

        self._close_icon = Gtk.Image.new_from_icon_name(
            "window-close-symbolic",
            Gtk.IconSize.LARGE_TOOLBAR,
        )
        close_btn = Gtk.Button()
        close_btn.set_image(self._close_icon)
        close_btn.set_relief(Gtk.ReliefStyle.NONE)
        close_btn.connect("clicked", self._on_close_clicked)

        title_label = Gtk.Label(label=guilabels.SPEECH_MONITOR)
        title_label.set_halign(Gtk.Align.CENTER)

        drag_bar = Gtk.EventBox()
        drag_bar.add(title_label)
        drag_bar.connect("button-press-event", self._on_drag_bar_press)

        close_row = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        close_row.pack_start(drag_bar, True, True, 0)
        close_row.pack_end(close_btn, False, False, 0)

        self._text_view = Gtk.TextView()
        self._text_view.set_editable(False)
        self._text_view.set_cursor_visible(True)
        self._text_view.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self._text_view.set_left_margin(6)
        self._text_view.set_right_margin(6)
        self._text_view.set_top_margin(6)
        self._text_view.set_bottom_margin(6)
        self._buffer = self._text_view.get_buffer()
        self._end_mark = self._buffer.create_mark("end", self._buffer.get_end_iter(), False)
        self._line_count = 0

        scrolled = Gtk.ScrolledWindow()
        scrolled.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled.add(self._text_view)

        border_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        border_box.pack_start(close_row, False, False, 0)
        border_box.pack_start(scrolled, True, True, 0)
        self.add(border_box)

        self.get_style_context().add_class("speech-monitor")
        titlebar.get_style_context().add_class("speech-monitor-titlebar")
        title_label.get_style_context().add_class("speech-monitor-title")
        close_btn.get_style_context().add_class("speech-monitor-close")
        border_box.get_style_context().add_class("speech-monitor-border")
        self._text_view.get_style_context().add_class("speech-monitor-text")

        if SpeechMonitor._shared_css_provider is None:
            SpeechMonitor._shared_css_provider = Gtk.CssProvider()
            Gtk.StyleContext.add_provider_for_screen(
                Gdk.Screen.get_default(),  # pylint: disable=no-value-for-parameter
                SpeechMonitor._shared_css_provider,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
            )
        self._apply_css()

        self.set_keep_above(True)
        self.set_skip_taskbar_hint(True)
        self.set_skip_pager_hint(True)

        self.set_focus(self._text_view)
        self.connect("notify::is-active", self._on_window_activated)
        self.connect("delete-event", self._on_delete_event)
        self.connect("key-press-event", self._on_key_press)

    def _on_window_activated(self, _window: Gtk.Window, _pspec: Any) -> None:
        """Grab focus on the text view whenever the window becomes active."""

        if self.is_active():
            self._text_view.grab_focus()

    def _on_close_clicked(self, _button: Gtk.Button) -> None:
        """Handle the close button click."""

        if self._on_close is not None:
            self._on_close()

    def _on_drag_bar_press(self, _widget: Gtk.Widget, event: Gdk.EventButton) -> bool:
        """Initiate a window move when the drag bar is clicked."""

        # pylint: disable=no-member
        if event.button == 1:
            self.begin_move_drag(event.button, int(event.x_root), int(event.y_root), event.time)
            return True
        return False

    def _on_delete_event(self, _window: Gtk.Window, _event: Gdk.Event) -> bool:
        """Intercept the close request to toggle the monitor off."""

        if self._on_close is not None:
            self._on_close()
        return True

    def _on_key_press(self, _window: Gtk.Window, event: Gdk.EventKey) -> bool:
        """Handle Ctrl+Plus/Minus/0 for zoom."""

        if not event.state & Gdk.ModifierType.CONTROL_MASK:
            return False

        if event.keyval in (Gdk.KEY_plus, Gdk.KEY_equal, Gdk.KEY_KP_Add):
            self._font_size = min(MAX_FONT_SIZE, self._font_size + ZOOM_STEP)
            self._apply_css()
            return True
        if event.keyval in (Gdk.KEY_minus, Gdk.KEY_KP_Subtract):
            self._font_size = max(MIN_FONT_SIZE, self._font_size - ZOOM_STEP)
            self._apply_css()
            return True
        if event.keyval in (Gdk.KEY_0, Gdk.KEY_KP_0):
            self._font_size = self._default_font_size
            self._apply_css()
            return True

        return False

    def set_font_size(self, size: int) -> None:
        """Updates the font size and reapplies CSS."""

        self._font_size = size
        self._apply_css()

    def reapply_css(
        self,
        foreground: str | None = None,
        background: str | None = None,
    ) -> None:
        """Reapplies CSS styling (e.g. after color changes)."""

        if foreground is not None:
            self._foreground = foreground
        if background is not None:
            self._background = background
        self._apply_css()

    def _apply_css(self) -> None:
        """Apply CSS styling for background color, text color, font size, and border."""

        bg = self._background
        fg = self._foreground
        close_px = max(16, self._font_size * 2)
        css = (
            f".speech-monitor {{ background-color: {bg}; }}\n"
            f".speech-monitor-titlebar {{ min-height: 0; padding: 0; margin: 0; "
            f"background-color: {bg}; border: none; box-shadow: none; }}\n"
            f".speech-monitor decoration {{ box-shadow: none; "
            f"border: none; }}\n"
            f".speech-monitor-title {{ color: {fg}; font-weight: bold; }}\n"
            f".speech-monitor-close {{ min-height: {close_px}px; min-width: {close_px}px; "
            f"padding: 2px; margin: 0; color: {fg}; }}\n"
            f".speech-monitor-close:hover {{ opacity: 0.7; }}\n"
            f".speech-monitor-border {{ border: 3px solid {fg}; }}\n"
            f".speech-monitor-text {{ font-size: {self._font_size}pt; }}\n"
            f".speech-monitor-text text {{ color: {fg}; background-color: {bg}; }}"
        )
        self._close_icon.set_pixel_size(close_px)
        if SpeechMonitor._shared_css_provider is not None:
            SpeechMonitor._shared_css_provider.load_from_data(css.encode())

    def clear(self) -> None:
        """Clears the speech monitor display."""

        self._buffer.set_text("")
        self._line_count = 0

    def _append_text(self, text: str) -> None:
        """Appends text, trims excess lines, and scrolls to the bottom."""

        self._buffer.insert(self._buffer.get_end_iter(), text)
        self._line_count += text.count("\n")
        if self._line_count > 500:
            excess = self._line_count - 500
            start = self._buffer.get_start_iter()
            cut_point = self._buffer.get_iter_at_line(excess)
            self._buffer.delete(start, cut_point)
            self._line_count = 500

        self._buffer.place_cursor(self._buffer.get_end_iter())
        self._text_view.scroll_mark_onscreen(self._end_mark)

    def write_text(self, text: str) -> None:
        """Appends spoken text to the monitor."""

        self._append_text(f"{text}\n")

    def write_key_event(self, key_description: str) -> None:
        """Appends a formatted key event entry to the monitor."""

        self._append_text(f"[{key_description}]\n")



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2018-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods

"""Provides support for synthesizing accessible input events."""

from __future__ import annotations

import time
from typing import NamedTuple

import gi

gi.require_version("Atspi", "2.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Atspi

from . import ax_device_manager, debug
from .ax_component import AXComponent
from .ax_object import AXObject
from .ax_text import AXText
from .ax_utilities import AXUtilities
from .ax_utilities_action import AXUtilitiesAction
from .ax_utilities_collection import AXUtilitiesCollection
from .ax_utilities_component import AXUtilitiesComponent
from .ax_utilities_debugging import AXUtilitiesDebugging
from .ax_utilities_role import AXUtilitiesRole
from .ax_utilities_text import CaretSetReason


class _Snapshot(NamedTuple):
    """Identifying info captured before a scroll so obj can be recovered if destroyed."""

    name: str
    text: str
    role: Atspi.Role
    parent_name: str
    parent_text: str
    parent_role: Atspi.Role

    @classmethod
    def from_object(cls, obj: Atspi.Accessible) -> _Snapshot:
        """Returns a snapshot describing obj and its parent."""

        parent = AXObject.get_parent(obj)
        return cls(
            name=AXObject.get_name(obj),
            text=AXText.get_all_text(obj),
            role=AXObject.get_role(obj),
            parent_name=AXObject.get_name(parent),
            parent_text=AXText.get_all_text(parent),
            parent_role=AXObject.get_role(parent),
        )


class AXEventSynthesizer:
    """Provides support for synthesizing accessible input events."""

    @staticmethod
    def _highest_ancestor(obj: Atspi.Accessible) -> bool:
        """Returns True if the parent of obj is the application or None."""

        parent = AXObject.get_parent(obj)
        return parent is None or AXUtilitiesRole.is_application(parent)

    @staticmethod
    def _is_scrolled_off_screen(
        obj: Atspi.Accessible,
        offset: int | None = None,
        ancestor: Atspi.Accessible | None = None,
    ) -> bool:
        """Returns true if obj, or the caret offset therein, is scrolled off-screen."""

        tokens = ["AXEventSynthesizer: Checking if", obj, "is scrolled offscreen"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        rect = AXComponent.get_rect(obj)
        ancestor = ancestor or AXUtilities.find_ancestor(obj, AXEventSynthesizer._highest_ancestor)
        if ancestor is None:
            tokens = ["AXEventSynthesizer: Could not get ancestor of", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        ancestor_rect = AXComponent.get_rect(ancestor)
        intersection = AXUtilitiesComponent.get_rect_intersection(ancestor_rect, rect)
        if AXUtilitiesComponent.is_empty_rect(intersection):
            tokens = ["AXEventSynthesizer:", obj, "is outside of", ancestor, ancestor_rect]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        if offset is None:
            tokens = ["AXEventSynthesizer:", obj, "is not scrolled offscreen"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        extents = AXText.get_character_rect(obj, offset)
        if AXUtilitiesComponent.is_empty_rect(extents):
            tokens = ["AXEventSynthesizer: Could not get character rect of", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        intersection = AXUtilitiesComponent.get_rect_intersection(extents, rect)
        if AXUtilitiesComponent.is_empty_rect(intersection):
            tokens = ["AXEventSynthesizer:", obj, "'s caret", extents, "not in obj", rect]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        return False

    @staticmethod
    def _generate_mouse_event(
        obj: Atspi.Accessible,
        relative_x: int,
        relative_y: int,
        event: str,
    ) -> bool:
        tokens = [
            "AXEventSynthesizer: Attempting to generate mouse event on",
            obj,
            f"at relative coordinates {relative_x},{relative_y}",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return ax_device_manager.get_manager().generate_mouse_event(
            obj,
            relative_x,
            relative_y,
            event,
        )

    @staticmethod
    def _mouse_event_on_character(obj: Atspi.Accessible, offset: int | None, event: str) -> bool:
        """Performs the specified mouse event on the current character in obj."""

        if offset is None:
            offset = max(AXText.get_caret_offset(obj), 0)

        if AXEventSynthesizer._is_scrolled_off_screen(obj, offset):
            AXEventSynthesizer.scroll_into_view(obj, offset)
            if AXEventSynthesizer._is_scrolled_off_screen(obj, offset):
                tokens = ["AXEventSynthesizer:", obj, "is still offscreen. Setting caret."]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                AXUtilities.set_caret_offset_with_reason(
                    obj, offset, CaretSetReason.SCROLL_INTO_VIEW
                )

        extents = AXText.get_character_rect(obj, offset)
        if AXUtilitiesComponent.is_empty_rect(extents):
            return False

        rect = AXComponent.get_rect(obj)
        intersection = AXUtilitiesComponent.get_rect_intersection(extents, rect)
        if AXUtilitiesComponent.is_empty_rect(intersection):
            tokens = ["AXEventSynthesizer:", obj, "'s caret", extents, "not in obj", rect]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        relative_x = (extents.x - rect.x) + extents.width / 2
        relative_y = (extents.y - rect.y) + extents.height / 2
        return AXEventSynthesizer._generate_mouse_event(obj, relative_x, relative_y, event)

    @staticmethod
    def _mouse_event_on_object(obj: Atspi.Accessible, event: str) -> bool:
        """Performs the specified mouse event on obj."""

        if AXEventSynthesizer._is_scrolled_off_screen(obj):
            AXEventSynthesizer.scroll_into_view(obj)
            if AXEventSynthesizer._is_scrolled_off_screen(obj):
                tokens = ["AXEventSynthesizer:", obj, "is still offscreen. Grabbing focus."]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                AXObject.grab_focus(obj)

        rect = AXComponent.get_rect(obj)
        relative_x = rect.width / 2
        relative_y = rect.height / 2
        return AXEventSynthesizer._generate_mouse_event(obj, relative_x, relative_y, event)

    @staticmethod
    def route_to_character(obj: Atspi.Accessible, offset: int | None = None) -> bool:
        """Routes the pointer to the current character in obj."""

        tokens = [f"AXEventSynthesizer: Attempting to route to offset {offset} in", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return AXEventSynthesizer._mouse_event_on_character(obj, offset, "abs")

    @staticmethod
    def route_to_object(obj: Atspi.Accessible) -> bool:
        """Moves the mouse pointer to the center of obj."""

        tokens = ["AXEventSynthesizer: Attempting to route to", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return AXEventSynthesizer._mouse_event_on_object(obj, "abs")

    @staticmethod
    def click_character(obj: Atspi.Accessible, offset: int | None = None, button: int = 1) -> bool:
        """Single click on the current character in obj using the specified button."""

        tokens = [f"AXEventSynthesizer: Attempting to click at offset {offset} in", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return AXEventSynthesizer._mouse_event_on_character(obj, offset, f"b{button}c")

    @staticmethod
    def click_object(obj: Atspi.Accessible, button: int = 1) -> bool:
        """Single click on obj using the specified button."""

        return AXEventSynthesizer._mouse_event_on_object(obj, f"b{button}c")

    @staticmethod
    def _scroll_to_location(
        obj: Atspi.Accessible,
        location: Atspi.ScrollType,
        start_offset: int | None = None,
        end_offset: int | None = None,
        root: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Attempts to scroll to the specified location; returns the scrolled accessible."""

        snapshot = _Snapshot.from_object(obj) if root is not None else None

        before = AXComponent.get_position(obj)
        AXText.scroll_substring_to_location(obj, location, start_offset, end_offset)
        AXObject.clear_cache(obj, False, "To obtain updated location after scroll.")
        after = AXComponent.get_position(obj)
        tokens = ["AXEventSynthesizer: Text scroll, before:", before, "after:", after]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if before == after:
            AXComponent.scroll_object_to_location(obj, location)
            AXObject.clear_cache(obj, False, "To obtain updated location after scroll.")
            after = AXComponent.get_position(obj)
            tokens = ["AXEventSynthesizer: Object scroll, before:", before, "after:", after]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if root is None or snapshot is None:
            return obj
        return AXEventSynthesizer._object_after_scroll(obj, root, snapshot)

    @staticmethod
    def _scroll_to_point(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        obj: Atspi.Accessible,
        x_coord: int,
        y_coord: int,
        start_offset: int | None = None,
        end_offset: int | None = None,
        root: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Attempts to scroll obj to the specified point; returns the scrolled accessible."""

        snapshot = _Snapshot.from_object(obj) if root is not None else None

        before = AXComponent.get_position(obj)
        AXText.scroll_substring_to_point(obj, x_coord, y_coord, start_offset, end_offset)
        AXObject.clear_cache(obj, False, "To obtain updated location after scroll.")
        after = AXComponent.get_position(obj)
        tokens = ["AXEventSynthesizer: Text scroll, before:", before, "after:", after]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if before == after:
            AXComponent.scroll_object_to_point(obj, x_coord, y_coord)
            AXObject.clear_cache(obj, False, "To obtain updated location after scroll.")
            after = AXComponent.get_position(obj)
            tokens = ["AXEventSynthesizer: Object scroll, before:", before, "after:", after]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if root is None or snapshot is None:
            return obj
        return AXEventSynthesizer._object_after_scroll(obj, root, snapshot)

    @staticmethod
    def scroll_into_view(
        obj: Atspi.Accessible,
        start_offset: int | None = None,
        end_offset: int | None = None,
        root: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Attempts to scroll obj into view; returns the scrolled accessible."""

        return AXEventSynthesizer._scroll_to_location(
            obj, Atspi.ScrollType.ANYWHERE, start_offset, end_offset, root
        )

    @staticmethod
    def scroll_to_center(
        obj: Atspi.Accessible,
        start_offset: int | None = None,
        end_offset: int | None = None,
        root: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Attempts to scroll obj to the center of its window; returns the scrolled accessible."""

        ancestor = AXUtilities.find_ancestor(obj, AXEventSynthesizer._highest_ancestor)
        if ancestor is None:
            tokens = ["AXEventSynthesizer: Could not get ancestor of", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return obj

        x_coord, y_coord = AXUtilitiesComponent.get_center_point(ancestor)
        return AXEventSynthesizer._scroll_to_point(
            obj, int(x_coord), int(y_coord), start_offset, end_offset, root
        )

    @staticmethod
    def scroll_to_top_edge(
        obj: Atspi.Accessible,
        start_offset: int | None = None,
        end_offset: int | None = None,
        root: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Attempts to scroll obj to the top edge; returns the scrolled accessible."""

        return AXEventSynthesizer._scroll_to_location(
            obj, Atspi.ScrollType.TOP_EDGE, start_offset, end_offset, root
        )

    @staticmethod
    def scroll_to_top_left(
        obj: Atspi.Accessible,
        start_offset: int | None = None,
        end_offset: int | None = None,
        root: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Attempts to scroll obj to the top left; returns the scrolled accessible."""

        return AXEventSynthesizer._scroll_to_location(
            obj, Atspi.ScrollType.TOP_LEFT, start_offset, end_offset, root
        )

    @staticmethod
    def scroll_to_left_edge(
        obj: Atspi.Accessible,
        start_offset: int | None = None,
        end_offset: int | None = None,
        root: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Attempts to scroll obj to the left edge; returns the scrolled accessible."""

        return AXEventSynthesizer._scroll_to_location(
            obj, Atspi.ScrollType.LEFT_EDGE, start_offset, end_offset, root
        )

    @staticmethod
    def scroll_to_bottom_edge(
        obj: Atspi.Accessible,
        start_offset: int | None = None,
        end_offset: int | None = None,
        root: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Attempts to scroll obj to the bottom edge; returns the scrolled accessible."""

        return AXEventSynthesizer._scroll_to_location(
            obj, Atspi.ScrollType.BOTTOM_EDGE, start_offset, end_offset, root
        )

    @staticmethod
    def scroll_to_bottom_right(
        obj: Atspi.Accessible,
        start_offset: int | None = None,
        end_offset: int | None = None,
        root: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Attempts to scroll obj to the bottom right; returns the scrolled accessible."""

        return AXEventSynthesizer._scroll_to_location(
            obj, Atspi.ScrollType.BOTTOM_RIGHT, start_offset, end_offset, root
        )

    @staticmethod
    def scroll_to_right_edge(
        obj: Atspi.Accessible,
        start_offset: int | None = None,
        end_offset: int | None = None,
        root: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Attempts to scroll obj to the right edge; returns the scrolled accessible."""

        return AXEventSynthesizer._scroll_to_location(
            obj, Atspi.ScrollType.RIGHT_EDGE, start_offset, end_offset, root
        )

    @staticmethod
    def _object_after_scroll(
        obj: Atspi.Accessible,
        root: Atspi.Accessible,
        snapshot: _Snapshot,
    ) -> Atspi.Accessible | None:
        """Returns obj or, if obj was destroyed by the scroll, a matching descendant of root."""

        def is_match(x: Atspi.Accessible) -> bool:
            if AXObject.is_dead(x):
                return False
            if AXObject.get_role(x) != snapshot.role:
                return False
            if AXObject.get_name(x) != snapshot.name:
                return False
            parent = AXObject.get_parent(x)
            if AXObject.get_role(parent) != snapshot.parent_role:
                return False
            if AXObject.get_name(parent) != snapshot.parent_name:
                return False
            if AXText.get_all_text(x) != snapshot.text:
                return False
            return AXText.get_all_text(parent) == snapshot.parent_text

        candidates: list[Atspi.Accessible] | None = None
        match = None
        for _attempt in range(3):
            time.sleep(0.05)
            if is_match(obj):
                return obj
            if AXObject.supports_collection(root):
                if candidates is None:
                    candidates = AXUtilitiesCollection.find_all_with_role(root, [snapshot.role])
                    msg = (
                        f"AXEventSynthesizer: {len(candidates)} role-matched candidates "
                        "for replacement search."
                    )
                    debug.print_message(debug.LEVEL_INFO, msg, True)
                match = next((c for c in candidates if is_match(c)), None)
            else:
                match = AXUtilities.find_descendant(root, is_match)
            if match is not None:
                break

        tokens = [
            "AXEventSynthesizer: object was destroyed during scroll; ",
            "matching descendant for",
            snapshot,
            "in",
            root,
            ":",
            match,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if match is not None:
            return match
        return obj if not AXObject.is_dead(obj) else None

    @staticmethod
    def try_all_clickable_actions(obj: Atspi.Accessible) -> bool:
        """Attempts to perform a click-like action if one is available."""

        actions = ["click", "press", "jump", "open", "activate"]
        for action in actions:
            if AXUtilitiesAction.do_named_action(obj, action):
                tokens = ["AXEventSynthesizer: '", action, "' on", obj, "performed successfully"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True

        if debug.debugLevel > debug.LEVEL_INFO:
            return False

        tokens = [
            "AXEventSynthesizer: Actions on",
            obj,
            ":",
            AXUtilitiesDebugging.actions_as_string(obj),
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return False

    @staticmethod
    def try_all_right_click_actions(obj: Atspi.Accessible) -> bool:
        """Attempts to perform a right-click-like action if one is available."""

        actions = ["show-context-menu", "menu"]
        for action in actions:
            if AXUtilitiesAction.do_named_action(obj, action):
                tokens = ["AXEventSynthesizer: '", action, "' on", obj, "performed successfully"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True

        if debug.debugLevel > debug.LEVEL_INFO:
            return False

        tokens = [
            "AXEventSynthesizer: Actions on",
            obj,
            ":",
            AXUtilitiesDebugging.actions_as_string(obj),
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return False


_synthesizer = AXEventSynthesizer()


def get_synthesizer() -> AXEventSynthesizer:
    """Returns the Event Synthesizer."""

    return _synthesizer



# Orca
#
# Copyright 2023-2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Wrapper for the Atspi.Action interface."""

from __future__ import annotations

import re

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi, GLib

from .ax_object import AXObject


class AXAction:
    """Wrapper for the Atspi.Action interface."""

    @staticmethod
    def get_n_actions(obj: Atspi.Accessible) -> int:
        """Returns the number of actions supported on obj."""

        if not AXObject.supports_action(obj):
            return 0

        try:
            count = Atspi.Action.get_n_actions(obj)
        except GLib.GError as error:
            msg = f"AXAction: Exception in get_n_actions: {error}"
            AXObject.handle_error(obj, error, msg)
            return 0

        return count

    @staticmethod
    def normalize_action_name(action_name: str) -> str:
        """Adjusts the name to account for differences in implementations."""

        if not action_name:
            return ""

        name = re.sub(r"(?<=[a-z])([A-Z])", r"-\1", action_name).lower()
        name = re.sub("[!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~]", "-", name)
        return name

    @staticmethod
    def get_action_name(obj: Atspi.Accessible, i: int, n_actions: int | None = None) -> str:
        """Returns the name of obj's action at index i."""

        if n_actions is None:
            n_actions = AXAction.get_n_actions(obj)
        if not 0 <= i < n_actions:
            return ""

        try:
            name = Atspi.Action.get_action_name(obj, i)
        except GLib.GError as error:
            msg = f"AXAction: Exception in get_action_name: {error}"
            AXObject.handle_error(obj, error, msg)
            return ""

        return AXAction.normalize_action_name(name)

    @staticmethod
    def get_action_description(obj: Atspi.Accessible, i: int, n_actions: int | None = None) -> str:
        """Returns the description of obj's action at index i."""

        if n_actions is None:
            n_actions = AXAction.get_n_actions(obj)
        if not 0 <= i < n_actions:
            return ""

        try:
            description = Atspi.Action.get_action_description(obj, i)
        except GLib.GError as error:
            msg = f"AXAction: Exception in get_action_description: {error}"
            AXObject.handle_error(obj, error, msg)
            return ""

        return description

    @staticmethod
    def get_action_localized_name(
        obj: Atspi.Accessible, i: int, n_actions: int | None = None
    ) -> str:
        """Returns the localized name of obj's action at index i."""

        if n_actions is None:
            n_actions = AXAction.get_n_actions(obj)
        if not 0 <= i < n_actions:
            return ""

        try:
            name = Atspi.Action.get_localized_name(obj, i)
        except GLib.GError as error:
            msg = f"AXAction: Exception in get_action_localized_name: {error}"
            AXObject.handle_error(obj, error, msg)
            return ""

        return name

    @staticmethod
    def get_action_key_binding(obj: Atspi.Accessible, i: int, n_actions: int | None = None) -> str:
        """Returns the key binding string of obj's action at index i."""

        if n_actions is None:
            n_actions = AXAction.get_n_actions(obj)
        if not 0 <= i < n_actions:
            return ""

        try:
            keybinding = Atspi.Action.get_key_binding(obj, i)
        except GLib.GError as error:
            msg = f"AXAction: Exception in get_action_key_binding: {error}"
            AXObject.handle_error(obj, error, msg)
            return ""

        # GTK4 does this.
        if keybinding == "<VoidSymbol>":
            return ""
        return keybinding

    @staticmethod
    def do_action(obj: Atspi.Accessible, i: int) -> bool:
        """Invokes obj's action at index i. The return value, if true, may be
        meaningless because most implementors return true without knowing if
        the action was successfully performed."""

        if not 0 <= i < AXAction.get_n_actions(obj):
            return False

        try:
            result = Atspi.Action.do_action(obj, i)
        except GLib.GError as error:
            msg = f"AXAction: Exception in do_action: {error}"
            AXObject.handle_error(obj, error, msg)
            return False

        return result
